unit FrmModuleU;


interface


uses
  Windows, Messages, SysUtils, Classes, Graphics, Forms, Menus,
  ShellAPI, MonitorU, LSAPI, Controls;


type
  TFrmModule = class(TForm)
    Popup: TPopupMenu;
    MnuOpen: TMenuItem;
    MnuEmpty: TMenuItem;
    MnuSep1: TMenuItem;
    MnuHide: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure MnuOpenClick(Sender: TObject);
    procedure FormDblClick(Sender: TObject);
    procedure FormClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure MnuRecycleClick(Sender: TObject);
    procedure MnuHideClick(Sender: TObject);
    procedure MnuToggleAniClick(Sender: TObject);
  private
    procedure TimRefreshTimer;
    procedure TimAnimateTimer;
    procedure SetAnimation(const Enable: boolean);
    // Window's messages
    procedure WMEraseBkgnd(var Message: TMessage); message WM_ERASEBKGND;
    procedure WMShowWindow(var Message: TMessage); message WM_SHOWWINDOW;
    procedure WMWindowPosChanging(var Message: TMessage); message WM_WINDOWPOSCHANGING;
    procedure WMDropFiles(var Message: TMessage); message WM_DROPFILES;
    procedure WMDisplayChange(var Message: TMessage); message WM_DISPLAYCHANGE;
    procedure WMTimer(var Message: TMessage); message WM_TIMER;
    // LiteStep's messages
    procedure LMGetRevID(var Message: TMessage); message LM_GETREVID;
  end;


const
  // The default filenames
  BackPixName: string = '.none';
  EmptyPixName: string = 'Recycler Empty.bmp';
  FullPixName: string = 'Recycler Full.bmp';
  AniName: string = '.none';

  // Revision strings
  rcsRevision: PChar = '$Revision: 4.0.4 $'#0;
  rcsID: PChar = '$Id: recycler.dpr,v 4.0.4 1999/11/01 17:53:00 calvert $'#0;

  IsAnimated: boolean = false; // Is set when the animation is running
  HasDockedBg: boolean = false; // If docked and true, we have the bg

  TimRefresh = 0; // Identifier for the refresh timer, MUST be 0
  TimAnimate = 1; // Identifier for the animation timer


var
  FrmModule: TFrmModule;
  BinMonitor: TMonitor;
  Transparent: boolean;
  Animated: boolean;
  State, AllPixOK, HideMenuHide: boolean;
  BackImage, FullImage, EmptyImage, AniImage, Snapshot: hBitmap;
  Frame: integer;
  AniDelay: integer;
  AniFrames: integer;
  AniCount: integer;
  AniLength: integer;
  FrameWidth, FrameHeight: integer;


implementation

{$R *.DFM}


uses
  LiteStep,
  Utilities;


/// Functions & Procedures /////////////////////////////////

/// Bang Commands ////////////////////////////////

// These are pretty self-explanatory

procedure BangRecyclerVisible(Wnd: hWnd; Args: PChar); cdecl;
var
  ShowMode: integer;
begin
  if Settings.DockedToWharf then exit;
  ShowMode := 0;
  if Args <> nil then
    if StrIComp(Args, 'true') = 0 then
      ShowMode := 1
    else if StrIComp(Args, 'false') = 0 then
      ShowMode := -1;
  with FrmModule do
    if ShowMode = 0 then
      Visible := not Visible
    else
      Visible := ShowMode = 1;
end;


procedure BangRecyclerEmpty(Wnd: HWnd; Args: PChar); cdecl;
begin
  BinMonitor.Suspend;
  FrmModule.SetAnimation(true);
  SHEmptyRecycleBin(Wnd, '', SHERB_NOCONFIRMATION + SHERB_NOPROGRESSUI);
  BinMonitor.Resume;
end;


procedure BangRecyclerOpen(Wnd: hWnd; Args: PChar); cdecl;
begin
  if not QuickExec('explorer.exe ::{645FF040-5081-101B-9F08-00AA002F954E}') then
    MessageBox(GetLiteStepWnd, 'Unable to open Recycler', 'Error', MB_ICONINFORMATION);
end;


/// Misc. Functions //////////////////////////////

//--------------------------------------------------
// Extract all settings from step.rc and modules.ini
//--------------------------------------------------
procedure GetSettings(var Settings: TSettings);
var
  szBuffer: array[0..1024] of char;
begin
  // These settings don't apply when docked
  if not Settings.DockedToWharf then with Settings do begin
    HideMenuHide := GetRCBool('RecyclerHideMenuHide', true);
    AlwaysOnTop := GetRCBool('RecyclerNotAlwaysOnTop', false);
    Hidden := GetRCBool('RecyclerStartHidden', true);
    Width := GetRCInt('RecyclerWidth', 64);
    Height := GetRCInt('RecyclerHeight', 64);
    Left := GetRCInt('RecyclerLeft', -Width);
    Top := GetRCInt('RecyclerTop', -Height);
  end;
  // These are always used
  Settings.DebugMode := GetRCBool('RecyclerDebugMode', true);
  GetRCString('RecyclerBack', szBuffer, pchar(BackPixName), SizeOf(szBuffer) - 1);
  BackPixName := szBuffer;
  GetRCString('RecyclerEmpty', szBuffer, pchar(EmptyPixName), SizeOf(szBuffer) - 1);
  EmptyPixName := szBuffer;
  GetRCString('RecyclerFull', szBuffer, pchar(FullPixName), SizeOf(szBuffer) - 1);
  FullPixName := szBuffer;
  GetRCString('RecyclerAnimation', szBuffer, pchar(AniName), SizeOf(szBuffer) - 1);
  AniName := szBuffer;
  AniDelay := GetRCInt('RecyclerAniDelay', 120);
  AniFrames := GetRCInt('RecyclerAniFrames', 4);
  AniLength := GetRCInt('RecyclerAniLength', 1200);

  SHERB_DebugMode := Settings.DebugMode;
  Transparent := LowerCase(BackPixName) = '.none';
  Animated := LowerCase(AniName) <> '.none';
  // Accept dropped files
  Settings.AcceptFiles := true;
end;


//-----------------------------------------
// Load all images that we are going to use
//-----------------------------------------
function LoadImages(const Transparent: boolean): boolean;
var
  bm: PBitmap;
begin
  Result := false;
  if not Transparent then
    BackImage := LoadLSImage(PChar(BackPixName), nil);
  EmptyImage := LoadLSImage(PChar(EmptyPixName), nil);
  FullImage := LoadLSImage(PChar(FullPixName), nil);
  if ((not Transparent) and (BackImage = 0)) or (EmptyImage = 0) or (FullImage = 0) then
    exit;
  if Animated then begin
    AniImage := LoadLSImage(PChar(AniName), 'RecyclerAnimation');
    if AniImage = 0 then
      exit;
    new(bm);
    GetObject(AniImage, SizeOf(bm^), bm);
    FrameWidth := bm.bmWidth;
    FrameHeight := bm.bmHeight div AniFrames;
    dispose(bm);
  end;
  Result := true;
end;


//------------------------------------------------------
// Converts edge relative X coordinates to absolute ones
//------------------------------------------------------
function AbsLeft(const Left: integer): integer;
begin
  Result := Left;
  if Result < 0 then
    Result := Result + GetSystemMetrics(SM_CXSCREEN);
end;


//------------------------
// Ditto for Y coordinates
//------------------------
function AbsTop(const Top: integer): integer;
begin
  Result := Top;
  if Result < 0 then
    Result := Result + GetSystemMetrics(SM_CYSCREEN);
end;


/// Methods ////////////////////////////////////////////////

/// FrmModule ////////////////////////////////////

//-------------------
// Mess up the system
//-------------------
procedure TFrmModule.FormCreate(Sender: TObject);
var
  szBuffer: array[0..1024] of char;
begin
  GetSettings(Settings);
  // Set the captions for the menu entries for localization
  GetRCString('RecyclerCapOpen', szBuffer, pchar(MnuOpen.Caption), SizeOf(szBuffer) - 1);
  MnuOpen.Caption := szBuffer;
  GetRCString('RecyclerCapEmpty', szBuffer, pchar(MnuEmpty.Caption), SizeOf(szBuffer) - 1);
  MnuEmpty.Caption := szBuffer;
  GetRCString('RecyclerCapHide', szBuffer, pchar(MnuHide.Caption), SizeOf(szBuffer) - 1);
  MnuHide.Caption := szBuffer;

  SHERB_Legacy := true;
  State := false;

  if Settings.DockedToWharf or HideMenuHide then begin
    MnuHide.Visible := false;
    MnuSep1.Visible := false;
  end;

  SetWindowPos(Handle, HWND_TOPMOST,
    AbsLeft(Settings.Left),
    AbsTop(Settings.Top),
    Settings.Width,
    Settings.Height,
    SWP_NOACTIVATE or SWP_NOZORDER
  );

  AllPixOK := LoadImages(Transparent);

  BinMonitor := TMonitor.Create(Handle, WM_TIMER, false, GetRCBool('RecyclerLegacyMode', true));
  BinMonitor.Priority := tpLower;

  AddBangCommand('!RecyclerEmpty', @BangRecyclerEmpty);
  AddBangCommand('!RecyclerOpen', @BangRecyclerOpen);
  AddBangCommand('!RecyclerVisible', @BangRecyclerVisible);
end;


//----------------------------------
// Clean up the mess we have created
//----------------------------------
procedure TFrmModule.FormDestroy(Sender: TObject);
begin
  KillTimer(Handle, TimAnimate);

  RemoveBangCommand('!RecyclerEmpty');
  RemoveBangCommand('!RecyclerOpen');
  RemoveBangCommand('!RecyclerVisible');

  if AllPixOK then begin
    DeleteObject(BackImage);
    DeleteObject(EmptyImage);
    DeleteObject(FullImage);
    DeleteObject(AniImage);
    DeleteObject(Snapshot);
  end;

  with BinMonitor do begin
    Terminate;
    WaitFor;
    Free;
  end;
end;


//-----------------
// Paint the module
//-----------------
procedure TFrmModule.FormPaint(Sender: TObject);
var
  StateImage: hBitmap;
  TempDC, FrameDC: hDC;
  bm: PBitmap;
  FrameBmp: hBitmap;
  CanvasDC: hDC;
  LastObjTemp: hGDIObj;
  LastObjFrame: hGDIObj;
begin
  // If not all images were found, draw something to indicate
  // our location and to tell the user that something is wrong
  if not AllPixOK then with Canvas do begin
    Brush.Style := bsClear;
    Pen.Width := 3;
    Pen.Color := clRed;
    MoveTo(0, 0);
    LineTo(Width, Height);
    MoveTo(Width, 0);
    LineTo(0, Height);
    exit;
  end;

  // To avoid being repetitititive
  if State then
    StateImage := FullImage
  else
    StateImage := EmptyImage;
  CanvasDC := GetDC(Handle);
  TempDC := CreateCompatibleDC(0);
  FrameDC := CreateCompatibleDC(0);
  FrameBmp := CreateCompatibleBitmap(CanvasDC, Width, Height);

  // IMPORTANT! When using SelectObject, it is VITAL that we save the previous
  // object, and restore it when we are done. This applies to temporary DCs as
  // well. If we fail to do this, the module leaks resources like a sieve.

  // Prepare the frame buffer
  LastObjTemp := SelectObject(TempDC, FrameBmp);
  LastObjFrame := SelectObject(FrameDC, Snapshot);
  BitBlt(TempDC, 0, 0, Width, Height, FrameDC, 0, 0, SRCCOPY);

  if IsAnimated then begin
    // Draw the animation frame on the buffer
    SelectObject(FrameDC, AniImage);
    TransparentBltLS(TempDC, 0, 0, FrameWidth, FrameHeight, FrameDC, 0, Frame * FrameHeight, RGB($FF, $00, $FF));
  end else begin
    // Draw the image on the buffer
    new(bm);
    SelectObject(FrameDC, StateImage);
    GetObject(StateImage, SizeOf(bm^), bm);
    with bm^ do
      TransparentBltLS(TempDC, 0, 0, bmWidth, bmHeight, FrameDC, 0, 0, RGB($FF, $00, $FF));
    dispose(bm);
  end;

  // Draw the frame buffer
  BitBlt(CanvasDC, 0, 0, Width, Height, TempDC, 0, 0, SRCCOPY);

  SelectObject(FrameDC, LastObjFrame);
  SelectObject(TempDC, LastObjTemp);

  DeleteObject(FrameBmp);
  DeleteDC(FrameDC);
  DeleteDC(TempDC);
  DeleteDC(CanvasDC);
end;


//-----------------------------------------------------------
// Force an update - useful when the wallpaper is changed, or
// the module is unable to automatically detect changes
//-----------------------------------------------------------
procedure TFrmModule.FormClick(Sender: TObject);
begin
  Hide;
  // The delay is needed so that the window beneath the module has time to refresh
  // I think I should make it a two pass routine, so that LS doesn't freeze...
  Sleep(20);
  Show;
  BinMonitor.Refresh := true;
end;


//-----------------------
// Redirect double-clicks
//-----------------------
procedure TFrmModule.FormDblClick(Sender: TObject);
begin
  MnuOpenClick(Sender);
end;


//-------------------
// Refresh the module
//-------------------
procedure TFrmModule.TimRefreshTimer;
begin
  State := BinMonitor.RecyclerFull;
  MnuEmpty.Enabled := State;
  if not (IsWindowVisible(Handle) and AllPixOK) then exit;
  Invalidate;
end;


//-------------------
// Animate the module
//-------------------
procedure TFrmModule.TimAnimateTimer;
begin
  if not (IsWindowVisible(Handle) and Animated) then exit;
  if AniCount <= 0 then
    SetAnimation(false);
  dec(AniCount);
  inc(Frame);
  if Frame >= AniFrames then
    Frame := 0;
  Paint;
end;


//-----------------------------
// Enable/disable the animation
//-----------------------------
procedure TFrmModule.SetAnimation(const Enable: boolean);
begin
  if (Enable = IsAnimated) or (not IsWindowVisible(Handle)) or (not Animated) then exit;
  if Enable then begin
    IsAnimated := true;
    SetTimer(Handle, TimAnimate, AniDelay, nil);
    AniCount := AniLength div AniDelay;
    Frame := 0;
  end else begin
    KillTimer(Handle, TimAnimate);
    IsAnimated := false;
  end;
  Invalidate;
end;

/// Message Handlers ///////////////////

//-------------------------------------------------------
// Defined to override the default behavior of the form,
// which is to erase the background when painting
// Since it is called, we can make it do something useful
//-------------------------------------------------------
procedure TFrmModule.WMEraseBkgnd(var Message: TMessage);
var
  CanvasDC: hDC;
  TempDC: hDC;
  BackDC: hDC;
  bm: PBitmap;

  LastObjTemp: hGDIObj;
  LastObjBack: hGDIObj;
begin
  // Create used structures
  CanvasDC := GetDC(Handle);
  TempDC := CreateCompatibleDC(0);
  BackDC := CreateCompatibleDC(0);
  // If Docked, save the bg once
  if Settings.DockedToWharf and not HasDockedBg then begin
    DeleteObject(Snapshot);
    Snapshot := CreateCompatibleBitmap(CanvasDC, Width, Height);
    LastObjTemp := SelectObject(TempDC, Snapshot);
    BitBlt(TempDC, 0, 0, Width, Height, CanvasDC, 0, 0, SRCCOPY);
    HasDockedBg := true;
  end else
    LastObjTemp := SelectObject(TempDC, Snapshot);
  // "Draw" the background
  if not (Transparent or IsAnimated) then begin
    new(bm);
    GetObject(BackImage, SizeOf(bm^), bm);
    LastObjBack := SelectObject(BackDC, BackImage);
    with bm^ do
      TransparentBltLS(TempDC, 0, 0, bmWidth, bmHeight, BackDC, 0, 0, RGB($FF, $00, $FF));
    SelectObject(BackDC, LastObjBack);
    dispose(bm);
  end;
  // Clean up
  SelectObject(TempDC, LastObjTemp);
  DeleteDC(BackDC);
  DeleteDC(TempDC);
  DeleteDC(CanvasDC);
  // To notify the system that we have "erased" the background
  Message.Result := -1;
end;


//----------------------------------------------------
// Grab the screen background when the window is shown
// This mustn't be done if docked though
//----------------------------------------------------
procedure TFrmModule.WMShowWindow(var Message: TMessage);
var
  CanvasDC: hDC;
  TempDC: hDC;
  LastObjTemp: hGDIObj;
begin
  if Settings.DockedToWharf or not BOOL(Message.wParam) then exit;
  CanvasDC := GetDC(Handle);
  TempDC := CreateCompatibleDC(0);
  DeleteObject(Snapshot);
  Snapshot := CreateCompatibleBitmap(CanvasDC, Width, Height);
  LastObjTemp := SelectObject(TempDC, Snapshot);
  BitBlt(TempDC, 0, 0, Width, Height, CanvasDC, 0, 0, SRCCOPY);
  SelectObject(TempDC, LastObjTemp);
  DeleteDC(TempDC);
  DeleteDC(CanvasDC);
end;

//----------------------------------------------------------------------
// If the module is somehow moved, it will keep its plaze in the z-order
//----------------------------------------------------------------------
procedure TFrmModule.WMWindowPosChanging(var Message: TMessage);
var
  WindowPos: ^TWindowPos;
begin
  WindowPos := pointer(Message.lParam);
  if not Settings.AlwaysOnTop then
    WindowPos^.flags := WindowPos^.flags or SWP_NOZORDER;
end;


//--------------------------------------------
// Handle files that are dropped on the module
//--------------------------------------------
procedure TFrmModule.WMDropFiles(var Message: TMessage);
var
  Aborted: boolean;
  FileNames: TStringList;
  FileString: string;
  LastWin: HWnd;
  i: longint;
  Undo: boolean;
begin
  BinMonitor.Suspend;
  SetAnimation(true);
  FileNames := GetDroppedFiles(HDROP(Message.WParam));
  DragFinish(HDROP(Message.WParam));
  FileString := '';
  for i := 0 to FileNames.Count - 1 do
    FileString := FileString + FileNames[i] + #13;
  LastWin := GetForegroundWindow;
  SetForegroundWindow(Handle);
  Undo := GetAsyncKeyState(VK_SHIFT) >= 0;
  QuickRecycle(GetLiteStepWnd, PChar(FileString), Undo, Aborted);
  SetForegroundWindow(LastWin);
  BinMonitor.Resume;
end;


//--------------------------------------------------------
// If negative coordinates are specified, this will ensure
// that the module sticks to them
//--------------------------------------------------------
procedure TFrmModule.WMDisplayChange(var Message: TMessage);
begin
  if Settings.DockedToWharf then exit;
  SetWindowPos(Handle, HWND_TOPMOST,
    AbsLeft(Settings.Left),
    AbsTop(Settings.Top),
    0,
    0,
    SWP_NOZORDER or SWP_NOSIZE or SWP_NOACTIVATE
  );
end;


//----------------------------------------------
// Call the functions associated with the timers
//----------------------------------------------
procedure TFrmModule.WMTimer(var Message: TMessage);
begin
  case Message.wParam of
    TimRefresh: TimRefreshTimer;
    TimAnimate: TimAnimateTimer;
  end;
end;


//---------------------------------------------------------------------
// Don't ask why we mangle the RevID this way, everyone else does it...
//---------------------------------------------------------------------
procedure TFrmModule.LMGetRevID(var Message: TMessage);
var
  Buf: PChar;
begin
  with Message do begin
    Buf := PChar(lParam);
    case wParam of
      0: begin
           StrCopy(Buf, 'recycler.dll: ');
           StrCat(Buf, @rcsRevision[11]);
           Buf[StrLen(Buf) - 1] := #0;
         end;
      1: begin
           StrCopy(Buf, @rcsID[1]);
           Buf[StrLen(Buf) - 1] := #0;
         end;
      else
        StrCopy(Buf, '');
    end;
    Result := StrLen(Buf);
  end;
end;


/// MenuItems ////////////////////////////////////

// As the bangs are already defined for these actions, just call them instead
// I won't even bother with commenting their purposes

procedure TFrmModule.MnuOpenClick(Sender: TObject);
begin
  BangRecyclerOpen(Handle, nil);
end;


procedure TFrmModule.MnuRecycleClick(Sender: TObject);
begin
  BangRecyclerEmpty(Handle, nil);
end;


procedure TFrmModule.MnuHideClick(Sender: TObject);
begin
  // More efficient than the bang
  if not Settings.DockedToWharf then Hide;
end;


procedure TFrmModule.MnuToggleAniClick(Sender: TObject);
begin
  SetAnimation(not IsAnimated);
end;


end.
