unit LiteStep;


interface


uses
  Windows,
  LSAPI,
  WharfData;


type
  TSettings = record
    LiteStepDir: string;
    AcceptFiles: boolean;
    AlwaysOnTop: boolean;
    DebugMode: boolean;
    DockedToWharf: boolean;
    Hidden: boolean;
    Left, Top, Width, Height: integer;
  end;


const
  LSDesktop = 'DesktopBackgroundClass';
  nMessages : array[0..1] of UINT = (LM_GETREVID, 0);


var
  Settings: TSettings;


function initModuleEx(ParentWnd: hWnd; DllInst: longint; Path: PChar): longint; cdecl;
function initModule(ParentWnd: hWnd; DllInst: longint; WD: WharfDataTypePtr): longint; cdecl;
procedure quitModule(DllInst: longint); cdecl;
function initWharfModule(ParentWnd: hWnd; DllInst: longint; WD: WharfDataTypePtr): longint; cdecl;
procedure quitWharfModule(DllInst: longint); cdecl;


exports
  initWharfModule index 1,
  quitWharfModule index 2,
  initModule      index 3,
  initModuleEx    index 4,
  quitModule      index 5;


implementation


uses
  FrmModuleU;


//-------------------------------
// To emulate the ? operator in C
//-------------------------------
function Select(const Cond: boolean; const A, B: integer): integer;
begin
  if Cond then Result := A else Result := B;
end;


//-------------------------------------------------
// Returns either LS's or explorer's desktop window
//-------------------------------------------------
function GetLiteStepDesktop: hWnd;
begin
  Result := FindWindow(LSDesktop, nil);
  if Result = 0 then
    Result := GetDesktopWindow;
end;


//---------------------------------------
// Called by LS when the module is loaded
//---------------------------------------
function initModuleEx(ParentWnd: hWnd; DllInst: longint; Path: PChar): longint;
var
  ParentRect: TRect;
begin
  // Initialize settings and create the module window
  with Settings do begin
    LiteStepDir := Path;
    if LiteStepDir[Length(LiteStepDir)] <> '\' then
      LiteStepDir := LiteStepDir + '\';
    if Settings.DockedToWharf then begin
      GetClientRect(ParentWnd, ParentRect);
      AlwaysOnTop := false;
      Left := 0;
      Top := 0;
      Width := ParentRect.Right;
      Height := ParentRect.Bottom;
      FrmModule := TFrmModule.CreateParented(ParentWnd)
    end else
      FrmModule := TFrmModule.Create(nil);
  end;
  // Setup the window
  with FrmModule do begin
    if not Settings.DockedToWharf then
      SetWindowLong(Handle, GWL_STYLE, WS_POPUP or WS_CLIPSIBLINGS or WS_CLIPCHILDREN);
    if not (Settings.AlwaysOnTop or Settings.DockedToWharf) then
      SetParent(Handle, GetLitestepDesktop);
    if Settings.AlwaysOnTop then
      SetWindowPos(Handle, HWND_TOPMOST,
        0,
        0,
        0,
        0,
        SWP_NOMOVE or SWP_NOSIZE or SWP_NOACTIVATE
      );
    SetWindowLong(Handle, GWL_USERDATA, magicDWord);
    SetWindowLong(Handle, GWL_EXSTYLE,
      GetWindowLong(Handle, GWL_EXSTYLE) or
      WS_EX_TOOLWINDOW or
      //Select(Settings.AlwaysOnTop, 0, WS_EX_TRANSPARENT) or
      Select(Settings.AcceptFiles, WS_EX_ACCEPTFILES, 0)
    );
    // Show the window, unless it is set to hidden
    ShowWindow(Handle, Select(Settings.Hidden, SW_HIDE, SW_SHOWNOACTIVATE));
    // Register with LS's message que
    SendMessage(GetLiteStepWnd, LM_REGISTERMESSAGE, Handle, LPARAM(@nMessages));
  end;

  Result := 1;
end;


//---------------------------------------------------------------------
// This is for compatibility with older releases, can safely be removed
//---------------------------------------------------------------------
function initModule(ParentWnd: hWnd; DllInst: longint; WD: WharfDataTypePtr): longint; cdecl;
begin
  Result := initModuleEx(ParentWnd, DllInst, WD^.lsPath);
end;


//------------------------------------------
// This is called when LS unloads the module
//------------------------------------------
procedure quitModule(DllInst: longint);
begin
  // Unregister from LS's message que
  SendMessage(GetLiteStepWnd, LM_UNREGISTERMESSAGE, FrmModule.Handle, LPARAM(@nMessages));
  // Destroy the window
  FrmModule.Free;
end;


//----------------------------------------------
// Called when the module is loaded by the wharf
//----------------------------------------------
function initWharfModule(ParentWnd: hWnd; DllInst: longint; WD: WharfDataTypePtr): longint;
begin
  Settings.DockedToWharf := true;
  Result := initModuleEx(ParentWnd, DllInst, WD^.lsPath);
end;


//-----------------------------------------
// Called when the wharf unloads the module
//-----------------------------------------
procedure quitWharfModule(DllInst: longint);
begin
  quitModule(DllInst);
end;


//--------------------
// Initialize settings
//--------------------
begin
  FillChar(Settings, SizeOf(Settings), 0);
end.
