unit MonitorU;


interface


uses
  Classes,
  Windows,
  Registry;


type
  TMonitor = class(TThread)
  private
    FLegacyMode: boolean;
    FRecyclerFull: boolean;
    FRefresh: boolean;
    FRegAvail: boolean;
    FWnd: hWnd;
    FMsg: integer;
    { Private declarations }
    NotifyEvent: THandle;
    Reg: TRegistry;
  public
    constructor Create(const OwnerWnd: hWnd; const Msg: integer; const CreateSuspended: boolean; const LegacyMode: boolean);
    destructor Destroy; override;
  protected
    property RegAvail: boolean read FRegAvail;
    procedure Execute; override;
  published
    property RecyclerFull: boolean read FRecyclerFull;
    property Refresh: boolean read FRefresh write FRefresh;
  end;


implementation


uses
  Utilities;


const
  HKCR_Path = '\CLSID\{645FF040-5081-101B-9F08-00AA002F954E}\DefaultIcon';
  HKCU_Path = '\Software\Classes\CLSID\{645FF040-5081-101B-9F08-00AA002F954E}\DefaultIcon';
  HKCU_Path2 = '\Software\Microsoft\Windows\CurrentVersion\Explorer\CLSID\{645FF040-5081-101B-9F08-00AA002F954E}\DefaultIcon';


//---------------------------------------------------------
// Setup the monitoring thread and create a registry object
//---------------------------------------------------------
constructor TMonitor.Create(const OwnerWnd: hWnd; const Msg: integer; const CreateSuspended: boolean; const LegacyMode: boolean);
begin
  inherited Create(CreateSuspended);
  // Setup property values
  FLegacyMode := LegacyMode;
  FRecyclerFull := false;
  FRefresh := true;
  FWnd := OwnerWnd;
  FMsg := Msg;
  // Setup the registry and terminate if entries aren't found
  Reg := TRegistry.Create;
  Reg.RootKey := HKEY_CURRENT_USER;
  FRegAvail := Reg.OpenKey(HKCU_Path2, false);
  if not FRegAvail then
    FRegAvail := Reg.OpenKey(HKCU_Path, false);
  if not FRegAvail then begin
    Reg.RootKey := HKEY_CLASSES_ROOT;
    FRegAvail := Reg.OpenKey(HKCR_Path, false);
  end;
  if RegAvail then
    FRecyclerFull := Reg.ReadString(#0) = Reg.ReadString('Full')
  else
    Terminate;
end;


//-------------------------------------------------------
// Shutdown the thread and dispose of the registry object
//-------------------------------------------------------
destructor TMonitor.Destroy;
begin
  try
    Reg.CloseKey;
  finally
    Reg.Free;
  end;
end;


//----------------------------
// The main loop of the thread
//----------------------------
procedure TMonitor.Execute;
begin
  // Initialization
  NotifyEvent := CreateEvent(nil, true, false, nil);
  RegNotifyChangeKeyValue(Reg.CurrentKey, false, REG_NOTIFY_CHANGE_LAST_SET, NotifyEvent, true);

  // Main loop
  while not Terminated do begin
    if WAIT_OBJECT_0 = WaitForSingleObject(NotifyEvent, 1000) then begin
      // For some reason you can't just reset this event
      CloseHandle(NotifyEvent);
      NotifyEvent := CreateEvent(nil, true, false, nil);
      RegNotifyChangeKeyValue(Reg.CurrentKey, false, REG_NOTIFY_CHANGE_LAST_SET, NotifyEvent, true);
      Refresh := true;
    end else
      if FLegacyMode then Refresh := FRecyclerFull <> (Reg.ReadString(#0) = Reg.ReadString('full'));
    if Refresh then begin
      FRecyclerFull := Reg.ReadString(#0) = Reg.ReadString('Full');
      if FMsg <> NULL then PostMessage(FWnd, FMsg, 0, 0);
      Refresh := false;
    end;
  end;
  // Cleanup
  CloseHandle(NotifyEvent);
end;



end.
