unit Utilities;

interface

uses
  SysUtils,
  Classes,
  Windows,
  ShellAPI,
  ShlObj,
  Registry;

type
  //-----------------------------------------
  // Wrapper for the SHFileOperation function
  //-----------------------------------------
  TFileOperation = class(TObject)
  private
    FFileOpStruct: TSHFileOpStruct;
  public
    constructor Create;
    procedure Clear;
    function Execute: integer;
  protected
  public
    property Wnd: HWnd read FFileOpStruct.Wnd write FFileOpStruct.Wnd;
    property wFunc: cardinal read FFileOpStruct.wFunc write FFileOpStruct.wFunc;
    property pFrom: PChar read FFileOpStruct.pFrom write FFileOpStruct.pFrom;
    property pTo: PChar read FFileOpStruct.pTo write FFileOpStruct.pTo;
    property fFlags: word read FFileOpStruct.fFlags write FFileOpStruct.fFlags;
    property fAnyOperationsAborted: longbool read FFileOpStruct.fAnyOperationsAborted;
    property hNameMappings: pointer read FFileOpStruct.hNameMappings write FFileOpStruct.hNameMappings;
    property lpszProgressTitle: PChar read FFileOpStruct.lpszProgressTitle write FFileOpStruct.lpszProgressTitle;
  end;


  TSHEmptyRecycleBin = function(Wnd: HWnd; LPCTSTR: PChar; DWORD: Word): integer; stdcall;


function ConvertDelimCR(const Source: string): string;
function GetDroppedFiles(Drop: HDrop): TStringList;
function GetFixedDrives: string;
function QuickRecycle(const Handle: HWnd; const Files: string; const Undo: boolean; var Aborted: boolean): integer;
function QuickExec(Command: string): boolean;

// Forgotten API functions :(
const
  SHERB_NOCONFIRMATION = $00000001;
  SHERB_NOPROGRESSUI   = $00000002;
  SHERB_NOSOUND        = $00000004;

  SHERB_DebugMode: boolean = false;
  SHERB_Legacy: boolean = false;


var
  SHEmptyRecycleBin: TSHEmptyRecycleBin;

implementation

const
  HKCR_Path = '\CLSID\{645FF040-5081-101B-9F08-00AA002F954E}\DefaultIcon';
  HKCU_Path = '\Software\Classes\CLSID\{645FF040-5081-101B-9F08-00AA002F954E}\DefaultIcon';
  RecycleErrorMsg = 'An exception was raised while emptying your recycle bin.'#13#10'please report the following error to the author.'#13#10#13#10;

var
  HShell32: THandle;

/// Forgotten API functions :(
// function SHEmptyRecycleBin; external 'SHELL32.DLL' name 'SHEmptyRecycleBinA';


/// TFileOperation ///////////////////////////////

//-----------------------------------
// Initialize the parameter structure
//-----------------------------------
constructor TFileOperation.Create;
begin
  inherited;
  Clear;
end;


//---------------------
// Reset all parameters
//---------------------
procedure TFileOperation.Clear;
begin
  FillChar(FFileOpStruct, sizeof(TSHFileOpStruct), 0);
end;


//----------------------
// Perform the operation
//----------------------
function TFileOperation.Execute: integer;
begin
  result := SHFileOperation(FFileOpStruct);
end;


/// File system //////////////////////////////////

//--------------------------------------------------------
// Replaces any occurence of the CR character with NUL 
// Used to format the parameter string for SHFileOperation
//--------------------------------------------------------
function ConvertDelimCR(const Source: string): string;
var
  i: longint;
begin
  result := Source + #13;
  for i := 0 to Length(result) do
    if PChar(result)[i] = #13 then PChar(result)[i] := #0;
end;


//--------------------------------------------------------------------
// Returns a list of strings containing the names of the dropped files
//--------------------------------------------------------------------
function GetDroppedFiles(Drop: HDrop): TStringList;
var
  Buffer: PChar;
  FileCount: longint;
  Size: longint;
  i: longint;
begin
  result := TStringList.Create;
  FileCount := DragQueryFile(Drop, $FFFFFFFF, nil, 0);
  for i := 0 to FileCount - 1 do begin
    Size := DragQueryFile(Drop, i, nil, 0) + 1;
    Buffer := StrAlloc(Size);
    DragQueryFile(Drop, i, Buffer, Size);
    result.Add(Buffer);
    StrDispose(Buffer);
  end;
end;


//------------------------------------------------------------
// Returns the drive letter for all fixed drives in the system
// Note that subst'ed drives are also returned
//------------------------------------------------------------
function GetFixedDrives: string;
var
  Drive: char;
begin
  result := '';
  Drive := 'A';
  while Drive <= 'Z' do begin
    if GetDriveType(PChar(Drive + ':\')) = DRIVE_FIXED then
      result := result + Drive;
    inc(Drive);
  end;
end;


//---------------------------------------------------------
// A shorthand function to erase files with SHFileOperation
//---------------------------------------------------------
function QuickRecycle(const Handle: HWnd; const Files: string; const Undo: boolean; var Aborted: boolean): integer;
var
  FileList: string;
  FileOperation: TFileOperation;
begin
  FileList := ConvertDelimCR(Files);
  FileOperation := TFileOperation.Create;
  with FileOperation do begin
    Wnd := Handle;
    wFunc := FO_DELETE;
    pFrom := PChar(FileList);
    fFlags := FOF_SILENT;
    if Undo then
      fFlags := fFlags or FOF_ALLOWUNDO;
    result := Execute;
    Aborted := fAnyOperationsAborted;
    Free;
  end;
end;


//--------------------------------------------
// As shorthand for the CreateProcess function
//--------------------------------------------
function QuickExec(Command: string): boolean;
var
  ProcInformation: TProcessInformation;
const
  StartupInfo: TStartupInfo = (
    cb: 0;
    lpReserved: nil; // unused
    lpDesktop: nil;
    lpTitle: nil;
    dwX: 0;
    dwY: 0;
    dwXSize: 300;
    dwYSize: 300;
    dwXCountChars: 0; // ignored
    dwYCountChars: 0; // ignored
    dwFillAttribute: 0; // ignored
    dwFlags: 0;
    wShowWindow: 0; // ignored
    cbReserved2: 0;
    lpReserved2: nil;
    hStdInput: 0; // ignored
    hStdOutput: 0; // ignored
    hStdError: 0; // ignored
  );
begin
  StartupInfo.cb := sizeof(TStartupInfo);
  result := CreateProcess(nil, PChar(Command), nil, nil, false, NORMAL_PRIORITY_CLASS, nil, nil, StartupInfo, ProcInformation);
end;


//-----------------------------------------------------
// A rather crude way of finding the recycle bin folder
//-----------------------------------------------------
function GetRecyclerPath(Drive: char): string;
begin
  Result := Drive + ':\Recycle';
  if FileExists(Result + 'd\Desktop.ini') then
    Result := Result + 'd\'
  else if FileExists(Result + 'r\Desktop.ini') then
    Result := Result + 'r\'
  else
    Result := '';
end;


//---------------------------------------------------
// An almost failsafe way of emptying the recycle bin
// without access to the SHEmptyRecycleBin function
//---------------------------------------------------
function EmptyRecycleBin(Wnd: HWnd; LPCTSTR: PChar; DWORD: Word): integer; stdcall;
var
  FileOperation: TFileOperation;
  Reg: TRegistry;
  BucketList: TStringList;
  DesktopIni: TextFile;
  Drive: char;
  Drives: string;
  Names: string;
  Flags: integer;
  i: integer;
  Error: integer;
begin
  BucketList := TStringList.Create;
  Reg := TRegistry.Create;
  FileOperation := TFileOperation.Create;
  result := 0;
  Error := 0;
  try
    with Reg do begin
      RootKey := HKEY_LOCAL_MACHINE;
      if OpenKey('\Software\Microsoft\Windows\CurrentVersion\explorer\BitBucket', false) then begin
        GetValueNames(BucketList);
        CloseKey;
      end;
    end;

    Drives := '';
    for i := 0 to BucketList.Count - 1 do begin
      if Length(BucketList[i]) <> 1 then continue;
      Drive := UpCase(PChar(BucketList[i])[0]);
      if Drive in ['A'..'Z'] then
        Drives := Drives + Drive;
    end;
    if SHERB_Legacy and (Drives = '') then
      Drives := GetFixedDrives;
    
    // We can't continue if we can't find any drives
    if Drives = '' then
      raise Exception.Create('No drives found.');

    Names := '';
    BucketList.Clear;
    for i := 0 to Length(Drives) - 1 do begin
      Drive := PChar(Drives)[i];
      if GetRecyclerPath(Drive) <> '' then begin
        BucketList.Add(GetRecyclerPath(Drive));
        Names := Names + GetRecyclerPath(Drive) + '*.*'#13;
      end;
    end;

    // We can't continue if we can't find the folder
    if Names = '' then
      raise Exception.Create('Couldn''t find the recycle bin.');

    // Delete all files in ?:\Recycled
    Names := ConvertDelimCR(Names);
    Flags := 0;
    if (DWORD and SHERB_NOCONFIRMATION) <> 0 then
      Flags := Flags + FOF_NOCONFIRMATION;
    if (DWORD and SHERB_NOPROGRESSUI) <> 0 then
      Flags := Flags + FOF_SILENT
    else
      Flags := Flags + FOF_SIMPLEPROGRESS;
    FileOperation.Wnd := Wnd;
    with FileOperation do begin
      wFunc := FO_DELETE;
      pFrom := PChar(Names);
      fFlags := Flags;
      Execute;
      
      // We mustn't continue the cleanup if we where interrupted
      if fAnyOperationsAborted then
        raise Exception.Create('Operation aborted.');
    end;

    // If not aborted, rewrite 'desktop.ini'
    for i := 0 to BucketList.Count - 1 do begin
      AssignFile(DesktopIni, BucketList[i] + 'Desktop.ini');
      Rewrite(DesktopIni);
      WriteLn(DesktopIni, '[.ShellClassInfo]');
      WriteLn(DesktopIni, 'CLSID={645FF040-5081-101B-9F08-00AA002F954E}');
      CloseFile(DesktopIni);
      FileSetAttr(BucketList[i] + 'Desktop.ini', faArchive or faHidden);
    end;

    // Cleanup registry
    with Reg do begin
      if OpenKey('\Software\Microsoft\Windows\CurrentVersion\explorer', false) then begin
        DeleteKey('BitBucket');
        CloseKey;
      end else
        Error := Error or 1;
      RootKey := HKEY_CURRENT_USER;
      if OpenKey(HKCU_Path, false) then begin
        WriteString('', ReadString('empty'));
        CloseKey;
      end else begin
        RootKey := HKEY_CLASSES_ROOT;
        if OpenKey(HKCR_Path, false) then begin
          WriteString('', ReadString('empty'));
          CloseKey;
        end else
          Error := Error or 2;
      end;
    end;
    // If the BitBucket wasn't found, we get confused
    if (Error and 1) <> 0 then
      raise Exception.Create('Couldn''t find ''BitBucket''.');
    // If we can't reset the icon, we get even more confused
    if (Error and 2) <> 0 then
      raise Exception.Create('Couldn''t reset icons.');
  except
    on E: Exception do begin
      result := -1;
      if SHERB_DebugMode then
        MessageBox(GetDesktopWindow, PChar(RecycleErrorMsg + E.Message), 'Unexpected event', MB_ICONINFORMATION);
    end;
  end;
  FileOperation.Free;
  Reg.Free;
  BucketList.Free;
end;



//----------------------------------------------------------------
// Find out if the system provides the SHEmptyRecycleBinA function
// and link to it if found, otherwise link to the legacy function
//----------------------------------------------------------------
initialization begin
  HShell32 := GetModuleHandle('Shell32.dll');
  if HShell32 <> 0 then begin
    @SHEmptyRecycleBin := GetProcAddress(HShell32, 'SHEmptyRecycleBinA');
    if @SHEmptyRecycleBin = nil then
      @SHEmptyRecycleBin := @EmptyRecycleBin;
  end;
end;

end.
